# Copyright 2020 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_concurrency import processutils

from neutron import privileged

COMMAND = 'ethtool'


@privileged.default.entrypoint
def offload(device, rx, tx, namespace=None):
    cmd = []
    if namespace:
        cmd += ['ip', 'netns', 'exec', namespace]
    rx = 'on' if rx else 'off'
    tx = 'on' if tx else 'off'
    cmd += [COMMAND, '--offload', device, 'rx', rx, 'tx', tx]
    return processutils.execute(*cmd)
